from unittest.mock import Mock, patch

import pytest

from nexrank.reranker import RerankResult, StructuredReranker


@pytest.fixture
def sample_documents():
    return [
        {
            "title": "Human Rights",
            "text": "Human rights are basic rights and freedoms.",
        },
        {
            "title": "Property Rights",
            "text": "Property rights are fundamental to ownership.",
        },
    ]


@pytest.fixture
def sample_query():
    return "What are human rights?"


@pytest.fixture
def reranker():
    return StructuredReranker()


def test_initialization():
    """Test reranker initialization with default values"""
    with patch("spacy.load") as mock_spacy_load:
        mock_nlp = Mock()
        mock_spacy_load.return_value = mock_nlp

        reranker = StructuredReranker()

        assert (
            reranker.cross_encoder.model_name == "cross-encoder/ms-marco-MiniLM-L-12-v2"
        )
        assert (
            reranker.bi_encoder.model_name == "sentence-transformers/all-MiniLM-L6-v2"
        )
        assert reranker.device in ["cuda", "cpu"]


def test_preprocess_text(reranker):
    """Test text preprocessing"""
    text = "This is a TEST sentence! With punctuation?"
    processed = reranker.preprocess_text(text)

    # Check if text is lowercase and punctuation is removed
    assert "!" not in processed
    assert "?" not in processed
    assert "TEST" not in processed
    assert all(c.islower() for c in processed if c.isalpha())


def test_rerank_empty_documents(reranker, sample_query):
    """Test reranking with empty document list"""
    result = reranker.rerank(sample_query, [])
    assert result == []


def test_compute_scores(reranker):
    """Test score computation for a document"""
    query = "human rights"
    doc = {"title": "Human Rights", "text": "Basic human rights and freedoms"}

    lexical_score, semantic_score = reranker.compute_scores(query, doc)

    # Check if scores are within expected range
    assert isinstance(lexical_score, float)
    assert isinstance(semantic_score, float)
    assert lexical_score >= 0
    assert semantic_score <= 1


def test_rerank_with_scores(reranker, sample_query, sample_documents):
    """Test reranking with score return"""
    results = reranker.rerank(sample_query, sample_documents, return_scores=True)

    assert isinstance(results, list)
    assert isinstance(results[0], RerankResult)

    # Check if results are sorted by score
    scores = [r.score for r in results]
    assert scores == sorted(scores, reverse=True)


def test_rerank_top_k(reranker, sample_query, sample_documents):
    """Test top_k parameter in reranking"""
    top_k = 1
    results = reranker.rerank(sample_query, sample_documents, top_k=top_k)

    assert len(results) == top_k


def test_rerank_result_format(reranker, sample_query, sample_documents):
    """Test the format of reranking results"""
    results = reranker.rerank(sample_query, sample_documents)

    assert isinstance(results, list)
    assert isinstance(results[0], dict)
    assert "title" in results[0]
    assert "text" in results[0]


def test_score_scaling(reranker, sample_query, sample_documents):
    """Test if scores are properly scaled"""
    results = reranker.rerank(sample_query, sample_documents, return_scores=True)

    scores = [r.score for r in results]
    # Check if scores are scaled between 0 and 1
    assert min(scores) >= 0
    assert max(scores) <= 1


def test_rerank_with_invalid_input(reranker):
    """Test handling of invalid input"""
    with pytest.raises(TypeError):
        reranker.rerank(None, [{"title": "test", "text": "test"}])

    with pytest.raises(KeyError):
        reranker.rerank("query", [{"invalid_key": "test"}])


@pytest.mark.parametrize(
    "top_k,expected_length",
    [
        (1, 1),
        (2, 2),
        (5, 2),  # Should return all documents if top_k > len(documents)
    ],
)
def test_rerank_different_top_k(
    reranker, sample_query, sample_documents, top_k, expected_length
):
    """Test reranking with different top_k values"""
    results = reranker.rerank(sample_query, sample_documents, top_k=top_k)
    assert len(results) == expected_length


def test_rerank_result_dataclass(reranker, sample_query, sample_documents):
    """Test the RerankResult dataclass functionality"""
    results = reranker.rerank(sample_query, sample_documents, return_scores=True)
    result = results[0]

    # Test all fields are present
    assert hasattr(result, "title")
    assert hasattr(result, "text")
    assert hasattr(result, "score")
    assert hasattr(result, "lexical_score")
    assert hasattr(result, "semantic_score")

    # Test to_dict method
    result_dict = result.to_dict()
    assert isinstance(result_dict, dict)
    assert "title" in result_dict
    assert "text" in result_dict



from dataclasses import dataclass
from typing import Dict, List, Union

import spacy
import torch
from rank_bm25 import BM25Okapi
from sentence_transformers import CrossEncoder, SentenceTransformer
from sklearn.preprocessing import MinMaxScaler
from tqdm import tqdm


@dataclass
class RerankResult:
    title: str
    text: str
    score: float
    lexical_score: float
    semantic_score: float

    def to_dict(self) -> Dict:
        """Convert result to dictionary format."""
        return {"title": self.title, "text": self.text}


class StructuredReranker:
    def __init__(
        self,
        cross_encoder_model: str = "cross-encoder/ms-marco-MiniLM-L-12-v2",
        bi_encoder_model: str = "sentence-transformers/all-MiniLM-L6-v2",
        device: str = None,
    ):
        self.device = device or ("cuda" if torch.cuda.is_available() else "cpu")
        self.cross_encoder = CrossEncoder(cross_encoder_model, device=self.device)
        self.bi_encoder = SentenceTransformer(bi_encoder_model, device=self.device)

        try:
            self.nlp = spacy.load("en_core_web_sm")
        except:
            print("Downloading spaCy model...")
            spacy.cli.download("en_core_web_sm")
            self.nlp = spacy.load("en_core_web_sm")

        self.scaler = MinMaxScaler()

    def compute_scores(self, query: str, doc: Dict[str, str]) -> tuple[float, float]:
        """Compute both lexical and semantic scores."""
        full_text = f"{doc['title']} {doc['text']}"

        # Compute lexical score
        processed_query = self.preprocess_text(query)
        processed_text = self.preprocess_text(full_text)
        query_tokens = processed_query.split()
        text_tokens = processed_text.split()
        bm25 = BM25Okapi([text_tokens])
        lexical_score = bm25.get_scores(query_tokens)[0]

        # Compute semantic score
        cross_score = self.cross_encoder.predict([(query, full_text)])
        query_embedding = self.bi_encoder.encode(query, convert_to_tensor=True)
        text_embedding = self.bi_encoder.encode(full_text, convert_to_tensor=True)
        bi_score = torch.nn.functional.cosine_similarity(
            query_embedding.unsqueeze(0), text_embedding.unsqueeze(0)
        ).item()

        semantic_score = 0.7 * float(cross_score) + 0.3 * bi_score
        return lexical_score, semantic_score

    def preprocess_text(self, text: str) -> str:
        """Clean and normalize text."""
        doc = self.nlp(text)
        tokens = [
            token.lemma_.lower()
            for token in doc
            if not token.is_stop and not token.is_punct
        ]
        return " ".join(tokens)

    def rerank(
        self,
        query: str,
        documents: List[Dict[str, str]],
        top_k: int = None,
        return_scores: bool = False,
    ) -> Union[List[Dict[str, str]], List[RerankResult]]:
        """
        Rerank documents and return in the same format as input.

        Args:
            query: Search query
            documents: List of dicts with 'title' and 'text' keys
            top_k: Number of top results to return
            return_scores: If True, return RerankResult objects with scores

        Returns:
            List of dicts in same format as input, or RerankResult objects if return_scores=True
        """
        if not documents:
            return []

        results = []
        for doc in tqdm(documents, desc="Reranking"):
            lexical_score, semantic_score = self.compute_scores(query, doc)
            final_score = 0.4 * lexical_score + 0.6 * semantic_score

            results.append(
                RerankResult(
                    title=doc["title"],
                    text=doc["text"],
                    score=final_score,
                    lexical_score=lexical_score,
                    semantic_score=semantic_score,
                )
            )

        # Sort by score
        results.sort(key=lambda x: x.score, reverse=True)

        # Scale scores
        scores = [[r.score] for r in results]
        scaled_scores = self.scaler.fit_transform(scores).flatten()
        for result, scaled_score in zip(results, scaled_scores):
            result.score = float(scaled_score)

        # Apply top_k
        if top_k:
            results = results[:top_k]

        # Return in requested format
        if return_scores:
            return results
        else:
            return [result.to_dict() for result in results]
